import trio
import asyncio

from ._base import BaseTrioEventLoop


class TrioEventLoop(BaseTrioEventLoop):
    """An asyncio event loop that runs on top of Trio, opened from
    within Trio code using :func:`open_loop`.
    """

    def _queue_handle(self, handle):
        self._check_closed()
        self._q_send.send_nowait(handle)
        return handle

    def default_exception_handler(self, context):
        """Default exception handler.

        This default handler simply re-raises an exception if there is one.

        Rationale:

        In traditional asyncio, there frequently is no context which the
        exception is supposed to affect.

        trio-asyncio, however, collects the loop and all its tasks in
        a Trio nursery. This means the context in which the error should
        be raised is known and can easily be controlled by the programmer.

        For maximum compatibility, this default handler is only used in
        asynchronous loops.

        """
        # TODO: add context.get('handle') to the exception

        exception = context.get('exception')
        if exception is None:
            message = context.get('message')
            if not message:
                message = 'Unhandled error in event loop'
            raise RuntimeError(message)
        else:
            raise exception

    def stop(self, waiter=None):
        """Halt the main loop.

        (Or rather, tell it to halt itself soon(ish).)

        :param waiter: an Event that is set when the loop is stopped.
        :type waiter: :class:`trio.Event`
        :return: Either the Event instance that was passed in,
                 or a newly-allocated one.

        """
        if waiter is None:
            if self._stop_wait is not None:
                return self._stop_wait
            waiter = self._stop_wait = trio.Event()
        else:
            if waiter.is_set():
                waiter.clear()

        def stop_me():
            waiter.set()
            raise StopAsyncIteration

        if self._stopped.is_set():
            waiter.set()
        else:
            self._queue_handle(asyncio.Handle(stop_me, (), self))
        return waiter

    def _close(self):
        """Hook for actually closing down things."""
        if not self._stopped.is_set():
            raise RuntimeError("You need to stop the loop before closing it")
        super()._close()



# This code implements basic asyncio compatibility

# Submodules are organized into the following layers, from highest to lowest;
# to avoid circular dependencies, submodules can only depend on other submodules
# in a strictly lower layer.
#
#                    nice facade: _adapter
# event loop dispatch and policy: _loop
#     event loop implementations: _async, _sync
#                event loop base: _base
#                      utilities: _handles, _util, _child, _deprecate, _version

from ._version import __version__  # noqa

from ._deprecate import TrioAsyncioDeprecationWarning
from ._util import run_aio_future, run_aio_generator
from ._base import BaseTrioEventLoop, TrioExecutor
from ._async import TrioEventLoop
from ._loop import (
    # main entry point:
    open_loop,
    # trio.run() + trio_asyncio.open_loop():
    run,
    # loop selection:
    current_loop,
    # forwarders to event loop methods:
    run_trio_task,
    run_trio,
    run_aio_coroutine,
)
from ._adapter import (
    aio_as_trio,
    trio_as_aio,
    # aliases for the above:
    asyncio_as_trio,
    trio_as_asyncio,
    # additional experimental goodie:
    allow_asyncio,
)

import importlib as _importlib
from . import _deprecate, _util

_deprecate.enable_attribute_deprecations(__name__)
__deprecated_attributes__ = {
    name: _deprecate.DeprecatedAttribute(
        _importlib.import_module("trio_asyncio._" + name.rstrip("_")),
        "0.11.0",
        issue=64,
        instead="an import from the top-level trio_asyncio package",
    )
    for name in ("adapter", "async_", "base", "child", "handles", "loop", "sync", "util")
}
__deprecated_attributes__.update(
    {
        name:
        _deprecate.DeprecatedAttribute(getattr(_loop, name), "0.11.0", issue=64, instead=None)
        for name in ("TrioPolicy", "TrioChildWatcher", "current_policy")
    }
)

# Provide aliases in the old place for names that moved between modules.
# Remove these when the non-underscore-prefixed module names are removed.
from . import _loop, _async
_async.open_loop = _loop.open_loop

_util.fixup_module_metadata(__name__, globals())



# This code implements helper functions that work without running
# a TrioEventLoop.

import trio
import asyncio
import sys
import outcome
import sniffio


async def run_aio_future(future):
    """Wait for an asyncio-flavored future to become done, then return
    or raise its result.

    Cancelling the current Trio scope will cancel the future.  If this
    results in the future resolving to an `asyncio.CancelledError`
    exception, the call to :func:`run_aio_future` will raise
    `trio.Cancelled`.  But if the future resolves to
    `~asyncio.CancelledError` when the current Trio scope was *not*
    cancelled, the `~asyncio.CancelledError` will be passed along
    unchanged.

    This is a Trio-flavored async function.

    """
    task = trio.lowlevel.current_task()
    raise_cancel = None

    def done_cb(_):
        trio.lowlevel.reschedule(task, outcome.capture(future.result))

    future.add_done_callback(done_cb)

    def abort_cb(raise_cancel_arg):
        # Save the cancel-raising function
        nonlocal raise_cancel
        raise_cancel = raise_cancel_arg
        # Attempt to cancel our future
        future.cancel()
        # Keep waiting
        return trio.lowlevel.Abort.FAILED

    try:
        res = await trio.lowlevel.wait_task_rescheduled(abort_cb)
        return res
    except asyncio.CancelledError as exc:
        if raise_cancel is not None:
            try:
                raise_cancel()
            finally:
                # Try to preserve the exception chain,
                # for more detailed tracebacks
                sys.exc_info()[1].__cause__ = exc
        else:
            raise


STOP = object()


async def run_aio_generator(loop, async_generator):
    """Return a Trio-flavored async iterator which wraps the given
    asyncio-flavored async iterator (usually an async generator, but
    doesn't have to be). The asyncio tasks that perform each iteration
    of *async_generator* will run in *loop*.
    """
    task = trio.lowlevel.current_task()
    raise_cancel = None
    current_read = None

    async def consume_next():
        t = sniffio.current_async_library_cvar.set("asyncio")
        try:
            item = await async_generator.__anext__()
            result = outcome.Value(value=item)
        except StopAsyncIteration:
            result = outcome.Value(value=STOP)
        except asyncio.CancelledError:
            # Once we are cancelled, we do not call reschedule() anymore
            return
        except Exception as e:
            result = outcome.Error(error=e)
        finally:
            sniffio.current_async_library_cvar.reset(t)

        trio.lowlevel.reschedule(task, result)

    def abort_cb(raise_cancel_arg):
        # Save the cancel-raising function
        nonlocal raise_cancel
        raise_cancel = raise_cancel_arg

        if not current_read:
            # There is no current read
            return trio.lowlevel.Abort.SUCCEEDED
        else:
            # Attempt to cancel the current iterator read, do not
            # report success until the future was actually cancelled.
            already_cancelled = current_read.cancel()
            if already_cancelled:
                return trio.lowlevel.Abort.SUCCEEDED
            else:
                # Continue dealing with the cancellation once
                # future.cancel() goes to the result of
                # wait_task_rescheduled()
                return trio.lowlevel.Abort.FAILED

    try:
        while True:
            # Schedule in asyncio that we read the next item from the iterator
            current_read = asyncio.ensure_future(consume_next())

            item = await trio.lowlevel.wait_task_rescheduled(abort_cb)

            if item is STOP:
                break
            yield item

    except asyncio.CancelledError as exc:
        if raise_cancel is not None:
            try:
                raise_cancel()
            finally:
                # Try to preserve the exception chain,
                # for more detailed tracebacks
                sys.exc_info()[1].__cause__ = exc
        else:
            raise


async def run_trio_generator(loop, async_generator):
    """Run a Trio generator from within asyncio"""
    while True:
        # Schedule in asyncio that we read the next item from the iterator
        try:
            item = await loop.trio_as_future(async_generator.__anext__)
        except StopAsyncIteration:
            break
        else:
            yield item


# Copied from Trio:
def fixup_module_metadata(module_name, namespace):
    seen_ids = set()

    def fix_one(qualname, name, obj):
        # avoid infinite recursion (relevant when using
        # typing.Generic, for example)
        if id(obj) in seen_ids:
            return
        seen_ids.add(id(obj))

        mod = getattr(obj, "__module__", None)
        if mod is not None and mod.startswith("trio_asyncio."):
            obj.__module__ = module_name
            # Modules, unlike everything else in Python, put fully-qualitied
            # names into their __name__ attribute. We check for "." to avoid
            # rewriting these.
            if hasattr(obj, "__name__") and "." not in obj.__name__:
                obj.__name__ = name
                obj.__qualname__ = qualname
            if isinstance(obj, type):
                for attr_name, attr_value in obj.__dict__.items():
                    fix_one(objname + "." + attr_name, attr_name, attr_value)

    for objname, obj in namespace.items():
        if not objname.startswith("_"):  # ignore private attributes
            fix_one(objname, objname, obj)
